import Endpoints from '@constants/Endpoints';
import Auth from '@packages/Auth';
import NetworkMockDriver from '@mocks/mock-network';
import StorageMockDriver from '@mocks/mock-storage';
import * as Network from '@libraries/Network';
import * as Storage from '@libraries/Storage';
import { SupportedChannel } from '@constants/Keys';
import MockProvider, {
  mockContext
} from '@packages/Auth/__mocks__/MockProvider.mock';
import waitty from '@utils/waitty';
import Shared from '@libraries/SharedContext';
// import getty from '@utils/getty';

beforeAll(() => {
  Network.setDriver(NetworkMockDriver);
  Storage.setDriver(StorageMockDriver);
});

function mockRenewSuccess() {
  // set request HTTP mock
  const tester = opt => {
    return opt.url.includes(
      Endpoints.api.v2.user.cognitoOpenIdOrRegister()
    );
  }
  const responder = () => ({
    items: {
      identityId: '$identityId',
      token: '$token',
      newcomer: 1
    },
    message: '成功',
    statusCode: 200
  });
  return NetworkMockDriver.mock(tester, responder);
}

function mockProfileSuccess() {
  // set request HTTP mock
  const tester = opt =>
    opt.url.includes(
      Endpoints.api.v2.user.profile()
    );
  const responder = () => ({
    items: {
      name: '$name',
      avatar: '$avatar'
    },
    message: '成功',
    statusCode: 200
  });
  return NetworkMockDriver.mock(tester, responder);
}

describe('General Auth tests', () => {
  const host = 'mock://auth.api';
  const auth = new Auth().host(host).channel(SupportedChannel.Driver).backoffBase(0);
  const mockProviders = [1, 2, 3].map(n => new MockProvider(n));
  mockRenewSuccess();
  mockProfileSuccess();

  it('connect providers should also initialize them, and ready will guarantee they been initialized', async () => {
    await auth.connectProvider(...mockProviders).ready();
    mockProviders.forEach((m, n) => {
      expect(m.getContext()).toEqual(mockContext(n + 1));
    });
  });

  it('sso provider state is consumed correctly, non-sso provider will not be initialized', async () => {
    const state = Date.now();
    const providers = [1, 2, 3].map(n => new MockProvider(n));
    await Storage.getDriver().setItem('anue.state', state);
    Shared.set('urld', {
      state,
      provider: providers[0].getName()
    });
    const auth2 = new Auth();
    await auth2.connectProvider(...providers).ready();
    // await waitty(1)
    console.log('examing ...');
    expect(providers[0].getContext()).toEqual(mockContext(1));
    expect(providers[1].getContext()).toBe(null);
    expect(providers[2].getContext()).toBe(null);
  });
});

describe('Auth events tests', () => {
  const host = 'mock://auth.api';
  const auth = new Auth().host(host).channel(SupportedChannel.Driver).backoffBase(0);
  const mockProviders = [4, 5, 6].map(n => new MockProvider(n));
  mockRenewSuccess();
  mockProfileSuccess();

  it('fires events when primary provider logged in', async () => {
    const listener = jest.fn().mockImplementation(e => {
      expect(e.source === mockProviders[0].getName())
    })

    auth.connectProvider(...mockProviders)
    auth.setSSOProvider(mockProviders[0].getName())

    auth.onStateChange(listener)
    mockProviders.forEach(m => {
      m.login()
    })
    await waitty(500)
    expect(listener).toBeCalledTimes(1)

    NetworkMockDriver.clearMocks()

  })

})
