/**
 * @anue-package Anue.Auth
 * @doc
 * Auth模組負責管理Token的交換、管理、刷新等環節。目前的架構上，由於使用者痛
 * 一使用[Cognito](https://aws.amazon.com/tw/cognito/)進行統一的身份識別因此不論是透過第三方的身份或是我們自己的
 * 認證機制（如Email）最後都得再透過API來取得Cognito ID和authorization token。
 *
 * ## 範例 (Social與Email登入)
 *
 * ```Javascript
 * import 'anue-fe-sdk'
 * import 'anue-fe-sdk/web-preset'
 * import Auth from 'anue-fe-sdk/Auth'
 * import GoogleWebProvider from 'anue-fe-sdk/GoogleWebProvider'
 * import FBWebProvider from 'anue-fe-sdk/FBWebProvider'
 * import EmailProvider from 'anue-fe-sdk/EmailProvider'
 * import Constants from 'anue-fe-sdk/Constants
 *
 * const google = new GoogleWebProvider({ apiKey, clientId })
 * const fb = new FBWebProvider({ clientId })
 * const email = new EmailProvider()
 *
 *
 * const auth = new Auth()
 *  // optional
 *  .channel(Constants.SupportedChannel.Stock)
 *  .onStateChange(ctx => {
 *    console.log('Auth state changed', ctx)
 *  })
 *
 * auth.connectProvider(google, fb, email)
 *
 *
 * ```
 */
import AbstractProvider from '@packages/Provider/AbstractProvider.abstract';
import Endpoints from '@constants/Endpoints';
import Keys, { SupportedChannel } from '@constants/Keys';
import Misc from '@constants/Misc';
import urly from '@utils/urly';
import * as Network from '@libraries/Network';
import * as Storage from '@libraries/Storage';
import getty from '@utils/getty';
import Eventy, { EventyHandler } from '@utils/eventy';
import Qable from '@utils/qable';
import { Web } from '@utils/isomorphy';
import waitty from '@utils/waitty';
import sso from '@utils/sso'

const refreshTokenEndpoints = [
  Endpoints.v1User.email.renew(),
  Endpoints.v1User.cognitoOpenIdOrRegister()
];

let backoff = -1;
let backoffExp = 0;
let backoffBase = 2

class Auth {

  static CacheTTL = 600000

  private mHost = Auth.getInitialURL();
  private mChannel: Anue.Auth.SupportedChannel = SupportedChannel.Unknown;
  private mIdToken: string | undefined;
  private mAccessToken: string | undefined;
  private mCognitoId: string | undefined;
  private mAuthorization: string | undefined;
  private mIsNewComer: boolean = false;
  private removeRequestInspector: (() => void) | null = null;
  private removeResponseInspector: (() => void) | null = null;
  private mConnectedProviders: Record<string, AbstractProvider> = {};
  private mIsReady: boolean = false;
  private mPrimaryProvider: AbstractProvider | null = null;
  private mProfile: Anue.Auth.UserProfile | null = null;
  private mSSOProvider: string = '';
  private mPortalUrl: string = 'https://login.cnyes.com';
  private mSyncLock: Record<string, boolean> = {};
  private mRenewRequestUrl: string = Endpoints.api.v2.user.cognitoOpenIdOrRegister();

  event: Eventy = new Eventy();

  constructor(config: Anue.Auth.Config = {}) {
    const { channel, host } = config;

    if (channel) {
      this.mChannel = channel;
    }
    if (host) {
      this.mHost = host;
    }
  }

  host = (host: string): Auth => {
    this.mHost = host;
    return this;
  };

  channel = (channel: SupportedChannel): Auth => {
    this.mChannel = channel;
    return this;
  };

  portal = (url: string): Auth => {
    this.mPortalUrl = url;
    return this;
  };

  backoffBase = (val: number): Auth => {
    backoffBase = val;
    return this
  }

  renewUrl = (url: string): Auth => {
    this.mRenewRequestUrl = url;
    return this;
  };

  launchPortal = async () => {
    const state = Date.now()
      .toString()
      .replace(/\d/, s => String.fromCharCode(+s + 97));
    Storage.getDriver().setItem('anue.state', state);
    location.href = `${this.mPortalUrl}?anue_auth_state=${state}&ref=${
      this.mChannel
    }`;
  };

  openPortal = (params: string) => {
    window.open(`${this.mPortalUrl}?${params}`)
  }

  ready = async (): Promise<Anue.Auth.Credentials | null> => {
    if (this.mIsReady) {
      return await this.getCredentialContext();
    }
    const value = await Storage.getDriver().getItem(
      Keys.Auth.Storage.SSOProvider
    );
    this.mSSOProvider = value;

    return new Promise<Anue.Auth.Credentials | null>(resolve => {
      const remove = this.event.on(Keys.Auth.Events.Loaded, () => {
        resolve(this.getCredentialContext());
        remove();
      });
    });
  };

  getPrimaryProvider = (): AbstractProvider | null => this.mPrimaryProvider;

  /**
   * @doc
   * @method onStateChange
   * @param fn EventyHandler<Anue.Auth.StateChangeEvent> 處理Auth狀態改變的函數，接受的參數是一個[Anue.Auth.StateChangeEvent]()物件
   * @returns [Auth](->Anue.Auth)
   * @public
   * 透過這個方法可以加入一個函數在Auth的狀態改變時被觸發，這也是應用端和Auth互動最主要的一個API。
   */
  onStateChange = (fn: EventyHandler<Anue.Auth.StateChangeEvent>): Auth => {
    this.event.on(Keys.Auth.Events.StateChange, fn);
    return this;
  };

  onError = (fn: EventyHandler<Anue.Auth.ErrorChangeEvent>) => {
    this.event.on(Keys.Auth.Events.Error, fn);
    return this;
  };

  @Web(Promise.resolve(null))
  private consumeState = async () => {

    // sso cookie is created by member center, which doesn't need a state verification
    const ssoCookieProvider = getty(sso.getSSOCookie(), ['provider'])

    console.log('[Auth] consume state .. sso=%o', ssoCookieProvider)
    if (ssoCookieProvider) {

      const storage = Storage.getDriver();
      this.mSSOProvider = ssoCookieProvider;
      await storage.setItem(Keys.Auth.Storage.SSOProvider, this.mSSOProvider);
      storage.setItem('anue.state', '');

    }

  };

  setSSOProvider = async (name: string) => {
    this.mSSOProvider = name;
    return this;
  };

  /**
   * @doc
   * @private
   * @method requestNetworkInspector
   * >! 這個方法不為外部所引用
   * 這個方法會在認證步驟`renew`或者`consume`完成之後自動由`Auth`添加至`Network`的
   * `requestInspector`佇列之中，此後每當有任何請求時`authorization`會被自動加入在
   * 請求的header中。
   */
  private requestNetworkInspector: Anue.Network.RequestInterceptor = async (
    options: Anue.Network.RequestOptions
  ) => {
    if (this.mAuthorization) {
      options.headers = options.headers || {};
      options.headers.authorization =
        options.headers.authorization || this.mAuthorization;
    }
    return await options;
  };

  /**
   * @doc
   * @private
   * @method responseNetworkInspector
   * >! 這個方法不為外部所引用
   * 這個方法會在認證步驟`renew`或者`consume`完成之後自動由`Auth`添加至`Network`的
   * `responseInspector`佇列之中，之後當請求回應狀態代碼為`401`時，這個方法會自動呼叫
   * [Auth.renew](#renew)並且試著將原本的請求**重送**。
   */
  private responseNetworkInspector: Anue.Network.ResponseInterceptor = async (
    options: Anue.Network.RequestOptions,
    response: Anue.Network.Response
  ) => {
    const { url } = options;
    const { statusCode } = response;
    const retries = getty(options, ['extra', 'resent']);
    const isNeedRefresh =
      statusCode === 401 &&
      // do not resend refresh token requests
      refreshTokenEndpoints.reduce((p, c) => p || url.includes(c), false) ===
        false;
    if (retries || !isNeedRefresh || !this.mPrimaryProvider) {
      return await response;
    }
    console.log('got 401 request options = %o', options);
    console.log(
      '%c[Auth] a 401 response captured, interceptor is trying to renew the credentials ..',
      'color: red;'
    );
    let credentials
    // try to refresh when 401 or refresh token failed
    try {
      credentials = await this.renew('Auth.Interceptor');
    }
    catch (err) {
      console.warn('[Auth] interceptor failed to renew token')
    }
    // if the credentials successfully updated, resend original request
    // otherwise just throw the error
    if (credentials) {
      options.extra = {
        ...options.extra,
        resent: true
      };
      response = await Network.getDriver().send(options);
    }

    return await response;
  };

  /**
   * @doc
   * @method getInitialURL
   * @private
   * 這個方法會依照當前的環境返回預設的登入驗證伺服器URL，在一般瀏覽器的環境下
   * 會返回當前的`hostname`作為預設的host來發出登入等相關的API
   * 在`React Native`或者是`NodeJS`這類的環境下預設會產生一個**空字串**
   */
  private static getInitialURL = (): string => {
    return typeof location === 'undefined'
      ? ''
      : location.protocol + '//' + location.host;
  };

  /**
   * @doc
   * @public
   * @promise
   * @method renew
   * @param ref? 用來追溯呼叫renew來源的參數，只有在log開啟的狀態下有用
   * @returns Anue.Auth.Credentials | null
   * 這個方法會先以假設使用者已經註冊過的情境來取得token，然後試著拿取使用者的`profile`
   *  - 當token取得失敗的時候就終止登入程序
   *  - 當取得的token無法拿到`profile`時，試著註冊使用者接著再嘗試取得`profile`，如果失敗則終止並清除登入資訊 (因為這個登入是無效的)
   */
  @Qable
  renew = async (ref?: any): Promise<Anue.Auth.Credentials | null> => {
    backoff = Date.now() + Math.pow(backoffBase, backoffExp) * 1000;

    // exponential backoff, prevent application layer calling this in high frequency
    if (Date.now() < backoff) {
      const waitFor = Math.pow(backoffBase, backoffExp++) * 1000;
      console.log('performing exponential backoff .. %s ms', waitFor);
      await waitty(waitFor);
    } else {
      backoff = -1;
      backoffExp = 1;
    }

    console.log('[Auth] renew called ref = %s', ref);

    if (!this.mPrimaryProvider) {
      console.log('[Auth] renew failed as it does not have any provider');
      return null;
    }

    const cachedUser = sso.getCachedUser()
    console.log('%c sso cached user = %o', 'color: brown', cachedUser)
    const useCache = (cachedUser && Date.now() - cachedUser.time < Auth.CacheTTL) as boolean

    let { context, profile } = await this.renewAndGetProfile(false, useCache ? cachedUser : null)

    if (!profile) {
      const retry = await this.renewAndGetProfile(true)
      context = retry.context
      profile = retry.profile
    }

    if (profile && context.token) {
      console.log('renew success profile=%o context=%o', profile, context)
      this.mProfile = profile
      return this.setCredentials(context).getCredentialContext();
    }

    return null

  };

  private renewAndGetProfile = async (retry: boolean = false, cachedUser: Anue.Auth.CachedUser = null): Promise<{
    context: APIResponse.v1.User.Token,
    profile: Anue.Auth.UserProfile | null
  }> => {

    let response: Anue.Auth.RefreshTokenResponse | null = null,
        profile: Anue.Auth.UserProfile | null = null;

    console.log('renew token with refresh = %o cachedUser = %o', retry, cachedUser)
    response = await this.renewRequest(retry);

    const token = getty(response, ['items', 'token']);
    if (!token) {
      throw {
        message: 'Failed to renew authorization token, invalid response.',
        detail: response
      }
    }

    console.log('[Auth] is trying to update user profile ... %o', response);

    if (cachedUser && cachedUser.cognitoId === getty(response, ['items', 'identityId'])) {
      profile = cachedUser.profile
    }
    else {
      profile = await this.getProfile({
        token: response.items.token,
        identityId: response.items.identityId,
        channel: this.mChannel
      });
    }

    if (retry && !profile) {
      this.clearCredentials()
      throw {
        status: 404,
        message: 'Failed to renew authorization token as the user profile does not exists'
      }
    }

    console.log('[Auth] latest profile is %o', profile);

    return { context: response.items, profile }
  }

  @Qable
  private renewRequest = async (
    registration: boolean = false
  ): Promise<Anue.Auth.RefreshTokenResponse> => {
    let endpoint, body;

    if (!this.mAccessToken) {
      throw 'auth.renew cannot be called without `accessToken`';
    }

    const network = Network.getDriver();
    const renewer = getty<() => Anue.Auth.RefreshTokenResponse>(
      this.mPrimaryProvider,
      ['renewer']
    );

    let response;

    // use renewer defined in provider instance
    if (renewer) {
      response = await renewer();
    } else {
      const isGoogleProvider = this.mIdToken;
      body = {
        type: isGoogleProvider ? 'google' : 'facebook',
        accessToken: this.mAccessToken || '',
        idToken: this.mIdToken || '',
        channel: this.mChannel,
        refresh: registration ? '0' : '1'
      } as APIRequest.v1.User.CognitoOpenIdOrRegister;

      console.log('%c[Auth] consuming .. %o', 'color:#c19774;', body);

      endpoint = this.mRenewRequestUrl;

      const url = urly(this.mHost, endpoint);
      const options = {
        method: Misc.RequestMethod.POST,
        url: url.toString(),
        body
      };
      response = await network.send<Anue.Auth.RefreshTokenResponse>(options);
    }

    return response;
  };

  @Qable
  private getProfile = async (
    args: Anue.Auth.GetProfileArgs
  ): Promise<Anue.Auth.UserProfile | null> => {
    if (!this.mPrimaryProvider) {
      console.log('[Auth] getProfile failed', null);
      return null;
    }

    return this.mPrimaryProvider.getProfile(args);
  };

  private initializeProviders = async (
    providers: AbstractProvider[],
    lazy: boolean = false
  ) => {
    console.log(
      '%c⚡️[Auth] connect providers %o lazy = %o',
      'font-weight:bolder;',
      providers,
      lazy
    );

    await this.consumeState();
    try {

      const providerCode = await Storage.getDriver().getItem(
        Keys.Auth.Storage.SSOProvider
      );

      if (providerCode) {
        this.mSSOProvider = providerCode;
      }
      const promises = providers.map(provider => {
        const providerName = provider.getName();
        this.mConnectedProviders[providerName] = provider;

        provider.onLoginStateChange(ctx => {
          this.syncProviderToken(provider, ctx);
        });

        // in lazy mode, sso provider should still be automatically initialized
        if (lazy && providerName !== this.mSSOProvider) {
          console.log(
            '%c[Auth] drop provider [%s] as it is not sso provider',
            'color:gray;',
            providerName
          );
          return Promise.resolve({})
        }

        return provider.init();
      });
      await Promise.all(promises)
      this.mIsReady = true;
      this.event.emit(Keys.Auth.Events.Loaded);
    }
    catch (err) {
      this.event.emit(Keys.Auth.Events.Error, err);
    }
  };

  lazyConnectProvider = (...providers: AbstractProvider[]): Auth => {
    this.initializeProviders(providers, true);
    return this;
  };

  /**
   * @doc
   * @name connectProvider
   * @param provider AbstractProvider 一個[Provider](->Anue.AbstractProvider)物件
   * 將這個provider物件的狀態同步連結到當前的Auth實體，這個行為會將[auth.consume](#consume)與該
   * provider之[login](->Anue.AbstractProvder#login)與[logout](->Anue.AbstractProvder#logout)方法進行連結。
   */
  connectProvider = (...providers: AbstractProvider[]): Auth => {
    this.initializeProviders(providers);
    return this;
  };

  private setCredentials = (data: APIResponse.v1.User.Token): Auth => {
    const { token, identityId, newcomer } = data;
    if (token && identityId && this.mPrimaryProvider) {
      this.mAuthorization = token;
      this.mCognitoId = identityId;
      this.mIsNewComer = newcomer === 1;
    }

    // add inspector for handling API authorization
    if (!this.removeRequestInspector) {
      this.removeRequestInspector = Network.addRequestInspector(
        this.requestNetworkInspector
      );
    }
    if (!this.removeResponseInspector) {
      this.removeRequestInspector = Network.addResponseInspector(
        this.responseNetworkInspector
      );
    }
    console.log(
      '%c[Auth] %cstate changed %conline',
      'font-weight:bolder;background-color: #b3ef99;',
      'background-color: #b3ef99;',
      'font-weight:bolder;background-color: #b3ef99'
    );

    this.event.emit(Keys.Auth.Events.StateChange, {
      status: 'online',
      source: this.mPrimaryProvider && this.mPrimaryProvider.getName(),
      context: this.getCredentialContext()
    });

    return this;
  };

  @Qable
  private syncProviderToken = async (
    provider: AbstractProvider,
    ctx: Anue.Provider.NullableResponse
  ) => {
    const providerName = provider.getName();
    console.log(
      '[Auth] login state change triggered by provider [%s] %o',
      providerName,
      ctx
    );
    if (this.mSSOProvider && provider.getName() !== this.mSSOProvider ) {
      console.log(
        '%c[Auth] renew aborted, as %s is not dedicated SSO provider.',
        'color: gray;',
        providerName
      );
    } else if (ctx && ctx.accessToken && !this.mSyncLock[providerName]) {
      console.log('provider [%s] sync locked');
      this.mSyncLock[providerName] = true;
      this.mAccessToken = ctx.accessToken;
      this.mIdToken = ctx.idToken || '';
      this.mPrimaryProvider = provider;
      try {

        const result = await this.renew(`Provider.${providerName}`);

        if (!result) {
          throw 'failed to renew'
        }

      } catch (err) {
        this.mSyncLock[providerName] = false;
        this.event.emit(Keys.Auth.Events.Error, err);
        console.log('provider [%s] sync unlocked due to an error %o', err);
        throw err;
      }
      console.log('provider [%s] sync unlocked');
    }
    else if (this.mSyncLock[providerName]) {
      console.log('[%s] sync operation aborted due to synclock');
    }
    else {
      this.event.emit(Keys.Auth.Events.StateChange, {
        status: 'offline',
        source: providerName
      });
    }
  };

  /**
   * @doc
   * @method clearCredentials
   * 清除所有儲存的認證資訊，等同於`logout`。當這個Auth物件有使用[connectProvider](->Anue.Auth#connectProvider)
   * 連結其他[Provider](->Anue.AbstractProvider)這個方法會在連結的Provider呼叫`logout`時一併被呼叫。
   */
  clearCredentials = async () => {
    this.mAuthorization = '';
    this.mIdToken = '';
    this.mAccessToken = '';
    this.mIdToken = '';
    this.mIsNewComer = false;
    this.mSSOProvider = '';
    this.mCognitoId = '';
    this.setSSOProvider('');

    // remove network auth inspector when credentials are cleared
    if (this.removeRequestInspector) {
      this.removeRequestInspector();
    }
    if (this.removeResponseInspector) {
      this.removeResponseInspector();
    }
    this.removeRequestInspector = null;
    this.removeResponseInspector = null;

    sso.removeSSOCookie()
    sso.removeCachedUser()

    console.log(
      '%c[Auth] %cstate changed %coffline',
      'font-weight:bolder;background-color: #fff693;',
      'background-color: #fff693;',
      'font-weight:bolder;background-color: #fff693;'
    );

    this.event.emit(Keys.Auth.Events.StateChange, {
      status: 'offline',
      source: this.mPrimaryProvider && this.mPrimaryProvider.getName()
    });
    return this
  };

  // alias method
  logout = this.clearCredentials;

  /**
   * Get last stored credential context object
   */
  getCredentialContext = (): Anue.Auth.Credentials | null => {
    return {
      authorization: this.mAuthorization,
      idToken: this.mIdToken,
      accessToken: this.mAccessToken,
      isNewComer: this.mIsNewComer,
      cognitoId: this.mCognitoId,
      profile: this.mProfile,
      ssoProvider: this.mSSOProvider
    };
  };
}

export default Auth;
