import AbstractProvider from '@packages/Provider/AbstractProvider.abstract';
import * as Storage from '@libraries/Storage';
import Misc from '@constants/Misc';
import Endpoints from '@constants/Endpoints';
import * as Network from '@libraries/Network';
import getty from '@utils/getty';
import { obfuscate, defuscate } from '@utils/obfuscator';
import urld from '@utils/urld';
import sso from '@utils/sso'

export interface IMailAndPassword {
  email: string;
  password: string;
}

export default class AnueMailProvider extends AbstractProvider {
  private static mHost =
    typeof location === 'undefined'
      ? ''
      : location.protocol + '//' + location.host;
  private mShouldStoreCredentials: boolean = urld.get<string>('onetime')
    ? false
    : true;

  getName = () => 'email';

  handleStateChangeImpl() {
    this.emitStateChange();
  }

  loginImpl = async (
    args?: IMailAndPassword
  ): Promise<Anue.Provider.NullableResponse> => {
    try {
      if (!args || !args.email || !args.password) {
        throw {
          message: `AnueMailProvider can not login with improper arguments "${JSON.stringify(
            args
          )}"`
        };
      }
      this.mContext = {
        accessToken: JSON.stringify(args),
        tokenType: ''
      };
    } catch (err) {
      console.warn(err);
      return await null;
    }
    const resp = await this.renewer();

    if (resp.statusCode !== 200) {
      throw resp;
    }

    return await this.mContext;
  };

  logoutImpl = async (): Promise<boolean> => {
    this.mContext = null;
    return await true;
  };

  refresh = this.loginImpl;

  renewer = async (): Promise<Anue.Auth.RefreshTokenResponse> => {
    const body = getty<string>(this.mContext, ['accessToken']);
    console.log('[AnueMailProvider] renew ..', body);
    if (!body) {
      throw {
        message:
          'AnueMailProvider unable to renew token as its context is empty'
      };
    }

    const resp = await Network.getDriver().send<
      APIResponse.Response<APIResponse.v1.User.Email.Renew>
    >({
      method: Misc.RequestMethod.POST,
      url: Endpoints.v1User.email.renew(),
      body: JSON.parse(body)
    });

    // store obfuscated password and email if it's valid
    if (resp.statusCode === 200 && this.mShouldStoreCredentials) {
      await Storage.getDriver().setItem(
        `anue-mail:${AnueMailProvider.mHost}`,
        obfuscate(body)
      );
    }

    return resp;
  };

  initImpl = async (): Promise<AbstractProvider> => {
    console.log(
      'get stored token from %s',
      `anue-mail:${location.protocol}//${location.host}`
    );
    const storage = Storage.getDriver();
    const ssoCookie = sso.getSSOCookie()
    const storedToken = await storage.getItem(
      `anue-mail:${location.protocol}//${location.host}`
    );
    const urlToken = urld.get<string>('mail_access_token') || null;
    let accessToken = ''

    if (ssoCookie && ssoCookie.provider === 'email' && ssoCookie.accessToken) {
      accessToken = ssoCookie.accessToken
    }

    console.log(
      '[AnueMailProvider] cookieToken = %s storedToken = %o urlToken =%o',
      accessToken,
      storedToken,
      urlToken
    );

    // hash token not found, try to find from storage
    if (!accessToken && urlToken) {
      accessToken = defuscate(urlToken)
      console.log(
        '%c[AnueMailProvider] consumed access token from url hash %o',
        'color: #0077ff;',
        this.mContext
      );
    } else if (!accessToken && storedToken) {
      try {
        console.log('[AnueMailProvider] defuscated stored token = %o ', accessToken);
        accessToken = defuscate(storedToken);
      } catch (err) {
        console.warn(err);
      }
    }

    if (accessToken) {
      this.mContext = {
        tokenType: '',
        accessToken
      }
      await this.renewer()
    }

    this.emitStateChange();

    return await this;
  };

  getContext(): Anue.Provider.NullableResponse {
    return this.mContext;
  }
}
