import AbstractProvider from '@packages/Provider/AbstractProvider.abstract';
import getty from '@utils/getty';
import Keys from '@constants/Keys';
import Rpc from '@utils/rpc';
import urld from '@utils/urld';
import SSO from '@utils/sso'

export default class FacebookOAuth2Provider extends AbstractProvider {
  private mIsInitialized = false;
  private mClientConfig: Anue.Provider.FacebookProviderConfig;
  private mRpc: Rpc | null = null;

  /**
   * @doc
   * @property mAuthUser
   * 這是一個相當重要的屬性，在Google Auth進行使用者登入時，需要知道要當入目前app的使用者是瀏覽器中的第幾號使用者
   * 如果這個欄位的資訊錯誤將會導致dpi frame在背景刷新時遭遇 `interaction_required` 的錯誤
   */
  private mAuthUser: string = '';

  static transform = <T = any>(obj): T => {
    const result = {};
    for (const prop in obj) {
      const match = prop.match(/[A-Z]?[a-z]+/g);
      if (match) {
        const key = match.join('_').toLowerCase();
        result[key] = obj[prop];
      }
    }
    return result as T;
  };

  constructor(config: Anue.Provider.FacebookProviderConfig) {
    super();
    if (!config.appId) {
      throw {
        message:
          'Failed to initialize FacebookOAuth2Provider without config.appId'
      };
    }
    this.mClientConfig = config;
  }

  getName = () => 'facebook';

  getType = () => Keys.SupportedProviderType.Social;

  getContext = (): Anue.Provider.Response | null => {
    return this.mContext;
  };

  /**
   * @doc
   * @method consume
   * @private
   * @async
   *
   * 這個方法通常會在登入導向的目標網站上進行作用，用來處理`location.hash`上帶的認證資訊，包含了`access_token`、
   * `token_type`、`authuser`等資訊。
   */
  private consume = async () => {
    const SSOCookie = SSO.getSSOCookie()
    const ssoProvider = urld.get('provider') || getty(SSOCookie, ['provider']);
    console.log('fb => %s', ssoProvider)

    if (ssoProvider === this.getName()) {
      const accessToken = urld.get<string>('access_token');
      const tokenType = urld.get<string>('token_type');

      this.mContext = { accessToken, tokenType };

    }
  };

  refresh = async (): Promise<Anue.Provider.Response> => {

    if (!this.mRpc) {
      this.mRpc = this.initializeRpc();
    }
    let accessToken = getty(this.mContext, ['accessToken']);
    let response;

    if (!accessToken) {
      response = await this.mRpc.rpcCall(
        'fbauth',
        encodeURIComponent(JSON.stringify(this.getRequestConfig()))
      );
      accessToken = response.accessToken;
    }

    if (accessToken) {
      this.mContext = { accessToken, tokenType: 'Bearer' };
      console.log(
        '%c[FacebookOAuth2Provider] refreshed token from rpc %o',
        'background-color: blue; font-weight: bolder; color: white',
        this.mContext
      );
    } else {
      throw {
        message: 'FacebookOAuth2Provider provider failed to refresh tokens',
        detail: this.mContext
      };
    }

    this.emitStateChange();

    return this.mContext;
  };

  private initializeRpc = (): Rpc => {
    const rpcUri = this.mClientConfig.rpcUri;

    if (!rpcUri) {
      throw {
        message:
          'FacebookOAuth2Provider cannot create rpc proxy without specifying `rpcUri` in the constructor'
      };
    }
    console.log('[FacebookOAuth2Provider] starting dpi on %s', rpcUri);

    return new Rpc('fbauth', rpcUri);
  };

  getRequestConfig = (isIdToken?: boolean) => {
    const configs: any = {
      client_id: this.mClientConfig.appId,
      redirect_uri: this.mClientConfig.rpcUri + '?dpi=fbauth',
      response_type: 'token'
    };
    if (isIdToken) {
      configs.nonce = Date.now();
      configs.prompt = 'none';
    }
    configs.authuser = this.mAuthUser;

    return configs;
  };

  initImpl = async (): Promise<AbstractProvider> => {
    if (this.mIsInitialized) {
      return this;
    }
    this.mIsInitialized = true;

    // consume tokens in url and `authuser` in storage
    await this.consume();

    if (this.mClientConfig.rpcUri) {
      this.mRpc = this.initializeRpc();
      await this.refresh();
    }

    this.emitStateChange();

    return this;
  };

  logoutImpl = async (): Promise<boolean> => {
    this.mContext = null
    return true;
  };

  loginImpl = async (args: { state: string }): Promise<Anue.Provider.NullableResponse> => {

    let url = 'https://www.facebook.com/v3.2/dialog/oauth' +
      `?client_id=${this.mClientConfig.appId}` +
      `&redirect_uri=${this.mClientConfig.redirectUri}` +
      '&response_type=token'

    if (args && args.state) {
      url += `&state=${args.state}`
    }

    location.href = url

    return null;
  };
}
