/**
 * @doc
 * @anue-provider anue.Providers.GoogleOAuth2Provider
 *
 * 有別於[GoogleWebProvider](anue.Providers.GoogleWebProvider)，這個類別是一個不使用Google JS SDK的實作
 * 驗證過程透過dpi frame和Google OAuth 2.0 REST API的流程將使用者導向不同的頁面進行使用者登入。
 *
 * > 使用GoogleOAuth2Provider必須注意的是，你必須在一個與當前應用網站相同的Origin嵌入fe-portal中的dpi.js以確保這個機制能正常運作
 */

import AbstractProvider from '@packages/Provider/AbstractProvider.abstract';
import * as Network from '@libraries/Network';
import getty from '@utils/getty';
import Keys from '@constants/Keys';
import Rpc from '@utils/rpc';
import urld from '@utils/urld';
import * as Storage from '@libraries/Storage';
import SSO from '@utils/sso'

export default class GoogleOAuth2Provider extends AbstractProvider {
  private mIsInitialized = false;
  private mClientConfig: Anue.Provider.GoogleProviderConfig;
  private mRpc: Rpc | null = null;

  /**
   * @doc
   * @property mAuthUser
   * 這是一個相當重要的屬性，在Google Auth進行使用者登入時，需要知道要當入目前app的使用者是瀏覽器中的第幾號使用者
   * 如果這個欄位的資訊錯誤將會導致dpi frame在背景刷新時遭遇 `interaction_required` 的錯誤
   */
  private mAuthUser: string = '';

  static transform = <T = any>(obj): T => {
    const result = {};
    for (const prop in obj) {
      const match = prop.match(/[A-Z]?[a-z]+/g);
      if (match) {
        const key = match.join('_').toLowerCase();
        result[key] = obj[prop];
      }
    }
    return result as T;
  };

  constructor(config: Anue.Provider.GoogleProviderConfig) {
    super();
    if (!config.clientId || !config.apiKey) {
      throw {
        message:
          'Failed to initialize GoogleProvider without config.appId and config.apiKey'
      };
    }
    this.mClientConfig = config;
  }

  getName = () => 'google';

  getType = () => Keys.SupportedProviderType.Social;

  getContext = (): Anue.Provider.Response | null => {
    return this.mContext;
  };

  /**
   * @doc
   * @method consume
   * @private
   * @async
   *
   * 這個方法通常會在登入導向的目標網站上進行作用，用來處理`location.hash`上帶的認證資訊，包含了`access_token`、
   * `token_type`、`authuser`等資訊。
   */
  private consume = async () => {
    const sso = SSO.getSSOCookie()
    const ssoProvider = urld.get('provider') || getty(sso, ['provider']);

    if (ssoProvider === this.getName()) {
      const accessToken = urld.get<string>('access_token');
      const tokenType = urld.get<string>('token_type');
      const authUser = urld.get<string>('authuser') || getty(sso, ['authuser']) || '0';
      const storage = Storage.getDriver();

      this.mContext = { accessToken, tokenType };

      await storage.setItem('anue:google-user', authUser);
    }
  };

  private recovery = async () => {
    const storage = Storage.getDriver();
    this.mAuthUser = (await storage.getItem('anue:google-user')) || '0';
  };

  refresh = async (): Promise<Anue.Provider.Response> => {
    if (!this.mRpc) {
      this.mRpc = this.initializeRpc();
    }

    let accessToken = getty(this.mContext, ['accessToken']);
    let idToken = getty(this.mContext, ['idToken']);
    let response;

    if (!accessToken) {
      response = await this.mRpc.rpcCall(
        'gauth',
        encodeURIComponent(JSON.stringify(this.getRequestConfig()))
      );
      accessToken = response.accessToken;
    }
    if (!idToken) {
      response = await this.mRpc.rpcCall(
        'gauth',
        encodeURIComponent(JSON.stringify(this.getRequestConfig(true)))
      );
      idToken = response.idToken;
    }

    if (idToken && accessToken) {
      this.mContext = { accessToken, idToken, tokenType: 'Bearer' };
      console.log(
        '%c[GoogleProvider] refreshed token from rpc %o',
        'background-color: blue; font-weight: bolder; color: white',
        this.mContext
      );
    } else {
      throw {
        message: 'google provider failed to refresh tokens',
        detail: this.mContext
      };
    }

    this.emitStateChange();

    return this.mContext;
  };

  private initializeRpc = (): Rpc => {
    const rpcUri = this.mClientConfig.rpcUri;

    if (!rpcUri) {
      throw {
        message:
          'GoogleOAuth2Provider cannot create rpc proxy without specifying `rpcUri` in the constructor'
      };
    }
    console.log('[GoogleProvider] starting dpi on %s', rpcUri);

    return new Rpc('gauth', rpcUri);
  };

  getRequestConfig = (isIdToken?: boolean) => {
    const configs: any = {
      client_id: this.mClientConfig.clientId,
      api_key: this.mClientConfig.apiKey,
      response_type: isIdToken ? 'id_token' : 'token',
      redirect_uri: this.mClientConfig.rpcUri + '?dpi=gauth',
      scope: [
        'openid',
        'email',
        'profile',
        'https://www.googleapis.com/auth/userinfo.profile',
        'https://www.googleapis.com/auth/userinfo.email',
        'https://www.googleapis.com/auth/plus.me'
      ].join(' ')
    };
    if (isIdToken) {
      configs.nonce = Date.now();
      configs.prompt = 'none';
    }
    configs.authuser = this.mAuthUser;

    return configs;
  };

  initImpl = async (): Promise<AbstractProvider> => {
    if (this.mIsInitialized) {
      return this;
    }
    this.mIsInitialized = true;

    // consume tokens in url and `authuser` in storage
    await this.consume();
    await this.recovery();

    try {
      if (this.mClientConfig.rpcUri) {
        this.mRpc = this.initializeRpc();
        await this.refresh();
      }
    } catch (err) {
      console.warn('[GoogleProvider] failed to grant credentials', err);
      this.mContext = null;
    }
    this.emitStateChange();

    return this;
  };

  logoutImpl = async (): Promise<boolean> => {
    const accessToken = getty<string>(this.mContext, ['accessToken']);

    if (accessToken) {
      await Network.getDriver().send({
        url: `https://accounts.google.com/o/oauth2/revoke?token=${accessToken}`
      });
      console.log('[GoogleProvider] revoked access token');
      this.mContext = null;
    }

    return true;
  };

  loginImpl = async (args: { state: string }): Promise<Anue.Provider.NullableResponse> => {
    const params: string[] = [];
    const configs = this.getRequestConfig();

    configs.prompt = 'select_account';
    configs.response_type = 'token';
    configs.redirect_uri = this.mClientConfig.redirectUri;
    configs.state = args.state
    delete configs.authuser;

    for (const c in configs) {
      params.push(`${c}=${decodeURIComponent(configs[c])}`);
    }

    location.href = `https://accounts.google.com/o/oauth2/v2/auth?${params.join(
      '&'
    )}`;

    return null;
  };
}
